﻿# api: ps
# type: functions
# title: clipboard
# description: Reads/writes to clipboard
# version: 0.9
# status: stable
# category: win32
# config: -
#
# Provides clipboard read and write shortcuts, with an HTML formatter.
#
# Additional callback for guimenu, to convert XAML/FlowDocument to HTML.
#

# syslib
Add-Type -AN System.Windows.Forms


#-- Send HTML formatted content to clipboard
function Set-ClipboardHtml($html) {
  #-- escape non-ASCII characters first
  $html = [regex]::replace($html, "([^\x01-\x7F])", {
      "&#" + ([int][char][string]$args[0]) + ";"
  })
  #-- wrap with HTML Format header
  $len = $html.length;  # section sizes
  $pfx = 190; $start = 165; $end = 35;
  $data = @"
Version:1.0
StartHTML:$(($pfx).toString().PadLeft(9, '0'))
EndHTML:$(($pfx+$start+$len+$end).toString().PadLeft(9, '0'))
StartFragment:$(($pfx+$start).toString().PadLeft(9, '0'))
EndFragment:$(($pfx+$start+$len).toString().PadLeft(9, '0'))
StartSelection:$(($pfx+$start).toString().PadLeft(9, '0'))
EndSelection:$(($pfx+$start+$len).toString().PadLeft(9, '0'))
SourceURL:http://localhost/#cmd-table
   <!DOCTYPE html><HTML> 
   <HEAD><TITLE>Clibboard</TITLE><META HTTP-EQUIV='Content-Type' CONTENT='text/html; charset=UTF-8'></HEAD>
   <BODY><!--Start-->      
$html
   <!--End--></BODY> </HTML>      
"@
  #-- send to system clipboard
  [System.Windows.Forms.Clipboard]::setData([System.Windows.Forms.Dataformats]::Html, $data)
}

#-- Write clipboard (text/plain)
function Set-Clipboard($text) {
    if ($text.length) {
        [void] [System.Windows.Forms.Clipboard]::setText($text)
    }
}

#-- Return current content (text/plain)
function Get-Clipboard() {
    return [System.Windows.Forms.Clipboard]::getText().trim()
}

<#
 #@tests
 return [System.Windows.Forms.Clipboard]::GetDataObject().getFormats()
#>



#-- extract FlowDocument as XAML, convert to HTML
function Convert-FlowdocumentToHtml($blocks) {
  <#
    .DESCRIPTION
       Hooked from guimenu ClipboardHtml button to generate HTML from current
       Output buffer.
    .NOTES
       - export Output FlowDocument to Xaml (not XamlPackage)
            <Section xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation" xml:space="preserve" TextAlignment="Left" Lin
            ontStretch="Normal" FontSize="12" Foreground="#FFFFFFFF" Typography.StandardLigatures="True" Typography.ContextualLigatu
            .StandardSwashes="0" Typography.ContextualSwashes="0" Typography.StylisticAlternates="0"><Paragraph Foreground="#FFFFFF0
            0" Margin="0,1,0,1" Background="#FF223388"><Run>ClickyColoury frontend to Multi-Tools  ?? ??? 2.5.1</Run></Paragraph><Pa
            ragraph Foreground="#FF88BB22" Margin="0,1,0,1"><Run> ??? New user interface, Powershell 5 compatible</Run></Paragraph><
            Paragraph Foreground="#FF88BB22" Margin="0,1,0,1"><Run> ??? Color clipboard not fully ported</Run></Paragraph></Section>
       - Then replace <Paragraph> and <Run>s with <p>/<br>
       - XML traversion makes it quite trivial:
            Foreground Margin  Background Run
            ---------- ------  ---------- ---
            #FFFFFF00  0,1,0,1 #FF223388  ClickyColoury frontend to Multi-Tools  ?? ??? 2.5.1
            #FF88BB22  0,1,0,1             ??? New user interface, Powershell 5 compatible
            #FF88BB22  0,1,0,1             ??? Color clipboard not fully ported
       - Paragraph may nest .Run and .Span/.Run tags
       - Only .Run usually contains actual text, but both.Span+.Para may set Background+Foreground colors
       - Substitute colors from dark background / white font, to white background and darker foreground colors
       - No explicit foreground color set for white inline <Runs> (just the default for FlowDocument widget)
       - ignore Margin attributes, as well as non-text nodes (no <Image/> tags without Format:XamlPackage anyhow)
  #>

    #-- read from `Output`
    try {
        # get FlowDoc as XAML
        $ms = New-Object System.IO.MemoryStream
        (Get-OutputTextRange).Save($ms, [System.Windows.DataFormats]::XAML);
        $xaml = [System.Text.Encoding]::UTF8.GetString($ms.getBuffer())
        $xaml = ($xaml -replace "[\x00-\x06]+","").trim()   # MemoryStream adds NUL padding sometimes
        #dbg
        #Write-Host $xaml
        #([xml]$xaml).Section.Paragraph | FT * | Out-String | Write-Host
        
        # traverse Paragraphs
        $html = ""
        ForEach ($p in ([xml]$xaml).Section.Paragraph) {
            $html += (Convert-SpanRunToHtml $p) + "<br>`r`n"
        }
        #Write-Host $html
        return $html
    }
    catch {
        $PSItem | FL * | Out-String | Write-Host
        $PSItem.Exception | FL * | Out-String | Write-Host
        $PSItem.Stacktrace | FL * | Out-String | Write-Host
    }
}
#-- Unpack nested .Span/.Run/#text inlines, inherit attributes to CSS colors
function Convert-SpanRunToHtml {
    Param(
        $p,
        $colors = @{},
        $map = @{   # WPF/light-on-dark to HTML/dark-on-light colors
            fg = @{
                "#ffffff00" = "#540" # Yellow
                "#ffff0000" = "#55070F" # Red
                "#ff0000ff" = "#114" # Blue
                "#ff008000" = "#040" # Green
                "#ff00ffff" = "#045" # Cyan
                "#ffffffff" = "#444" # White
                "#ff808080" = "#666" # Gray
                "#ff000000" = "#222" # Black 
                "#ff9988dd" = "#543" # .header
            }
            bg = @{
                "#ffffff00" = "#f3f3aa" # Yellow
                "#ffff0000" = "#ef8a77" # Red
                "#ff0000ff" = "#a5a5f5" # Blue
                "#ff008000" = "#aea" # Green
                "#ff00ffff" = "#a1ece5" # Cyan
                "#ffffffff" = "#999" # White
                "#ff808080" = "#bbb" # Gray
                "#ff000000" = "#222" # Black 
                "#ff102070" = "#ccf" # .header
            }
            html = @{
                "<" = "&lt;"
                ">" = "&gt;"
                '✔' = '<font color=#151>✔</font>'
                '✘' = '<font color=#712>✘</font>'
            }
        }
    )
    #-- literal content
    if ($p -is [string]) {
        # html escaping, etc.
        $p = [regex]::replace($p, $map.html.keys -join "|", {
            Param($m); $map.html[$m.Groups[0].Value]
        })
        # whitespace
        if ($p -match "^\s{2,}|\s{3,}") {
            $colors["white-space"] = "pre"
        }
        $p = $p -replace "\r*\n","<br>`r`n"

        # add style=
        if ($colors) {
            $style = (@($colors.keys) | % { $_+":"+$colors[$_] }) -join "; "
            return "<span style='$style'>$p</span>"
        }
        else {
            return $p
        }
    }
    #-- color mapping
    if ($c = $p.Foreground) {
        $colors["color"] = if ($map.fg[$c]) { $map.fg[$c] } else { Convert-FlowDocColor $c 0.6 }
    }
    if ($c = $p.Background) {
        $colors["background"] = if ($map.bg[$c]) { $map.bg[$c] } else { Convert-FlowDocColor $c 1.5 }
    }
    if ($p.FontWeight -eq "Bold") {
        $colors["font-weight"] = "bold"
    }
    #-- inline traversal
    if ($p.Span) {
        return ($p.Span | % { Convert-SpanRunToHtml ($_) (@{}+$colors) }) -join ""
    }
    elseif ($p.Run) {
        return ($p.Run | % { Convert-SpanRunToHtml ($_) ($colors) }) -join ""
    }
    elseif ($p."#text") {
        return Convert-SpanRunToHtml ($p."#text") ($colors)
    }
    else {
        #$p | FT | Out-String | Write-Host
    }
}
#-- up/down hexcolors a bit
function Convert-FlowDocColor {
    Param($hex, $quant = 1.3)
    if ($hex -match "#\w\w(\w\w)(\w\w)(\w\w)") {
        return "#" + (($matches[1..3] | % { [convert]::ToInt32($_, 16) } | % { $_ * $quant } | % { [convert]::ToString($_, 16).PadLeft(2, 48) }) -join "")
    }
}

